/**
 * This is a simple program to test XIM
 * Gaspar Sinai <gsinai@yudit.org>
 */
#include <stdio.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/X.h>
#include <X11/Xatom.h>
#include <X11/Xproto.h>
#include <X11/keysymdef.h>
#include <X11/Xutil.h>
#include <X11/Xlocale.h>


static char*inputMethod = "interxim";
static char*xlocale = "";

static void handleEvents (Display* display, Window window);
static XIC* switchInputMethod (Display* display, Window w, int on);
static void destroyIMCB (XIM xim, XPointer p0, XPointer p1);

static int events = \
    ExposureMask \
  | StructureNotifyMask \
  | SubstructureNotifyMask \
  | ButtonPressMask \
  | ButtonReleaseMask \
  | Button1MotionMask \
  | Button2MotionMask \
  | Button3MotionMask \
  | KeyPressMask \
  | KeyReleaseMask \
  | KeymapStateMask \
  | PropertyChangeMask \
  | EnterWindowMask \
  | LeaveWindowMask \
  | FocusChangeMask;

Atom wmProtocols, wmDeleteWindow;

int
main (int argc, char* argv[])
{
  int screen; 
  Window root;
  int black, white;
  int x=100;
  int y=100;
  unsigned int width = 400;
  unsigned int height = 300;
  Window window;
  char * n = "X11 Test";
  XClassHint    classHints;
  XWMHints wmHints;
  Display* display = XOpenDisplay (0);
  int i;
  for (i=0; i<argc; i++)
  {
    if (!strcmp (argv[i], "-im") && i+1 < argc)
    {
      i++;
      inputMethod = argv[i];
    }
    else if (!strcmp (argv[i], "-locale") && i+1 < argc)
    {
      i++;
      xlocale = argv[i];
    }
  }
  fprintf (stderr, "Current Options: \n");
  fprintf (stderr, "  -im '%s'\n", inputMethod);
  fprintf (stderr, "  -locale '%s'\n", xlocale);
  fprintf (stderr, "Hot keys:\n"); 
  fprintf (stderr, "  i - start im\n"); 
  fprintf (stderr, "  o - stop im\n"); 
  fprintf (stderr, "  q - quit\n"); 
  fprintf (stderr, "Shift-Space starts and ends conversion.\n"); 

  if (display == 0)
  {
    fprintf (stderr, "can not open display\n");
    return 1;
  }
  /* screen */
  screen = DefaultScreen (display);
  root = RootWindow (display, screen);
  white = WhitePixel (display, screen);
  black = BlackPixel (display, screen);
  wmProtocols = XInternAtom (display, "WM_PROTOCOLS", False);
  wmDeleteWindow = XInternAtom (display, "WM_DELETE_WINDOW", False);
  window = XCreateSimpleWindow (display, root, x, y,
       width, height, 0, white, black);
  classHints.res_name =  n;
  classHints.res_class = n;
  wmHints.initial_state = NormalState;
  wmHints.input = True;
  wmHints.flags = StateHint | InputHint;
  XSetWMProperties (display, window, 0, 0, 0, 0, 0, &wmHints, &classHints);
  XSetWMProtocols (display, window, &wmDeleteWindow, 1);
  XStoreName (display, window, n);
  XMapRaised (display, window);
  XSelectInput (display, window, events);
  handleEvents (display, window);
  return 0;
}

static XIC icnull = 0;

/*
 * key i opens "inputMethod" input method.
 * key o closes "inputMethod" input method.
 */
static
void handleEvents (Display* display, Window window)
{
  XEvent event;
  XIC* icres = &icnull;
  while (1)
  {
    XNextEvent (display, &event);
    if (XFilterEvent (&event, event.xany.window))
    //if (XFilterEvent (&event, None))
    {
      fprintf (stderr, "Event filtered out\n");
      continue;
    }
    if (event.type == KeyPress)
    {
      Status status_return;
      char asci[32];
      KeySym key = 0;
      int cnt;
      XIC ic = (window == event.xany.window) ? *icres : 0;
#ifdef X_HAVE_UTF8_STRING
      if (ic == 0)
      {
        cnt = XLookupString (&event.xkey, asci, sizeof (asci), &key, 0);
      }
      else
      {
        cnt = Xutf8LookupString (ic, &event.xkey,
               asci, sizeof (asci), &key, &status_return);
        /* interxim  */
      }
#else
      cnt = XLookupString (&event.xkey, asci, sizeof (asci), &key, 0);
#endif
      fprintf (stderr, "keypress... cnt=%d", cnt);
      if (cnt == 0 && (key & 0xff000000) == 0x01000000)
      {
        fprintf (stderr, " ucs4=%X.\n", (key & 0x00ffffff));
      }
      else if (cnt == 0)
      {
        fprintf (stderr, "\n");
        continue;
      }
      else
      {
        fprintf (stderr, "\n");
      }
      if (window == event.xany.window && asci[0] == 'i' || asci[0] == 'I')
      {
         icres = switchInputMethod (display, window, 1);
      }
      else if (window == event.xany.window && asci[0] == 'o' || asci[0] == 'O')
      {
         icres = switchInputMethod (display, window, 0);
      }
      else if (window == event.xany.window && asci[0] == 'q' || asci[0] == 'Q')
      {
         fprintf (stderr, "Quit...\n");
         return;
      }
    }
    else if (event.type ==  MapNotify)
    {
      if (event.xany.window == window)
      {
         XSetInputFocus (display, window, RevertToNone, CurrentTime);
      }
    }
    else if (event.type ==  ClientMessage)
    {
      if ((Atom)event.xclient.message_type == wmProtocols
        && (Atom)event.xclient.data.l[0] == wmDeleteWindow
        && event.xclient.format == 32)
      {
         fprintf (stderr, "Exiting...\n");
         return;
      }
    }
  }
}

typedef struct {
  XIC ic;
  XIM im;
  int ison;
  char *name;
} InputContext;

static char iarr[64];


static XIC*
switchInputMethod (Display* display, Window window, int on)
{
  static InputContext input = {0, 0, 0, ""}; 
  char* p=0;
  XIMCallback  imCallback;
  XIMStyles* xim_styles;
  int  xim_ev_mask = 0;
  int i;

  if (on == input.ison) return &icnull;
  fprintf (stderr, "Switching %s method %s\n", inputMethod, on?"on":"off");
  input.ison = on;
  if (!input.ison)
  {
     if (input.ic != 0) XDestroyIC (input.ic);
     if (input.im != 0) XCloseIM (input.im);
     input.ic = 0;
     input.im = 0;
     input.name = "";
     return &icnull;
  }
  if (!setlocale (LC_ALL, xlocale))
  {
    fprintf (stderr, "Locale %s is not supported by C library.\n", xlocale);
    return &icnull;
  }
  if (!XSupportsLocale())
  {
    fprintf (stderr, "Locale %s is not supported by X11 library.\n", xlocale);
    return &icnull;
  }
  strcpy (iarr, "@im=");
  strncat (iarr, inputMethod, sizeof (iarr)-5);
  iarr[sizeof (iarr)-1] = 0;

  p = XSetLocaleModifiers (iarr);
  if (p==0 || *p == 0)
  {
    fprintf (stderr, "setting modifiers '%s' failed.\n", inputMethod);
    return &icnull;
  }
  input.name = inputMethod;
  input.im = XOpenIM(display, 0, 0, 0);
  if (input.im == 0)
  {
    fprintf (stderr, "can not open input method %s.\n", inputMethod);
    return &icnull;
  }
  imCallback.client_data = (char*)&input;
  imCallback.callback = (XIMProc) destroyIMCB;
  XSetIMValues (input.im, XNDestroyCallback , &imCallback , 0);
  /* get styles */

  if (XGetIMValues(input.im, XNQueryInputStyle, &xim_styles, 0)
      || xim_styles==0 || xim_styles->count_styles==0)
  {
    fprintf (stderr, "input method doesn't support any style\n");
    XCloseIM (input.im); input.im = 0;
    return &icnull;
  }
  /* get root style */
  for (i=0; i<xim_styles->count_styles; i++)
  {
    if (xim_styles->supported_styles[i] ==
     (XIMPreeditNothing  | XIMStatusNothing))
    {
       break; 
    }
  }
  if (i==xim_styles->count_styles)
  {
    fprintf (stderr, "input method does not have root style.\n");
    XFree(xim_styles);
    XCloseIM (input.im); input.im = 0;
    return &icnull;
  }
  XFree(xim_styles);
  input.ic = XCreateIC(input.im, 
       XNInputStyle, XIMPreeditNothing |XIMStatusNothing,
       XNClientWindow, window,
       XNFocusWindow, window,
       0);
  if (input.ic == 0)
  {
    fprintf (stderr, "can not create input context .\n");
    XCloseIM (input.im); input.im = 0;
    return &icnull;
  }
  XGetICValues (input.ic , XNFilterEvents , &xim_ev_mask , NULL);
  xim_ev_mask |= KeyPressMask|KeyReleaseMask;
  XSetICValues (input.ic , XNFilterEvents , &xim_ev_mask , NULL);

  xim_ev_mask |= events;
  //This will result in a bad window attribute error
  XSelectInput (display, window, xim_ev_mask);
  XSetICFocus (input.ic);
  return &input.ic;
}

/**
 * I tested it - never worked - gaspar
 */
static void
destroyIMCB (XIM xim, XPointer p0, XPointer p1)
{
  InputContext* icx = (InputContext*) p0; 
  fprintf (stderr, "X Input server crashed %s.\n", icx->name);
  icx->im = 0;
  icx->ic = 0;
  icx->ison = 0;

}

